@@ -1,4 +1,5 @@
 # Copyright (C) 2014-2018 Enzien Audio, Ltd.
+# Copyright (C) 2023-2024 Wasted Audio
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -13,11 +14,13 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-from collections import Counter
-from collections import OrderedDict
 import decimal
 import os
 import re
+from collections import Counter
+from collections import OrderedDict
+from pathlib import Path
+from typing import List, Dict, Optional, Type, Any, Generator
 
 from .HeavyObject import HeavyObject
 from .HeavyGraph import HeavyGraph              # pre-converted Heavy graphs
@@ -43,9 +46,11 @@ from .NotificationEnum import NotificationEnum
 class PdParser:
 
     # library search paths
+    __LIB_DIR = os.path.join(os.path.dirname(__file__), "libs")
     __HVLIB_DIR = os.path.join(os.path.dirname(__file__), "libs", "heavy")
     __HVLIB_CONVERTED_DIR = os.path.join(os.path.dirname(__file__), "libs", "heavy_converted")
     __PDLIB_DIR = os.path.join(os.path.dirname(__file__), "libs", "pd")
+    __ELSELIB_DIR = os.path.join(os.path.dirname(__file__), "libs", "else")
     __PDLIB_CONVERTED_DIR = os.path.join(os.path.dirname(__file__), "libs", "pd_converted")
 
     # detect a dollar argument in a string
@@ -54,40 +59,41 @@ class PdParser:
     # detect width parameter e.g. "#X obj 172 79 t b b, f 22;"
     __RE_WIDTH = re.compile(r", f \d+$")
 
-    def __init__(self):
+    def __init__(self) -> None:
         # the current global value of $0
         # Note(joe): set a high starting value to avoid potential user naming conflicts
         self.__DOLLAR_ZERO = 1000
 
         # a counter of all Pd objects in the graph
-        self.obj_counter = Counter()
+        self.obj_counter: Counter = Counter()
 
         # search paths at this graph level
-        self.__search_paths = []
+        self.__search_paths: List = []
 
     @classmethod
-    def get_supported_objects(clazz):
+    def get_supported_objects(cls) -> List:
         """ Returns a set of all pd objects names supported by the parser.
         """
-        pd_objects = [os.path.splitext(f)[0] for f in os.listdir(clazz.__PDLIB_DIR) if f.endswith(".pd")]
-        pd_objects.extend(PdParser.__PD_CLASSES.keys())
+        pd_objects = [os.path.splitext(f)[0] for f in os.listdir(cls.__PDLIB_DIR) if f.endswith(".pd")]
+        pd_objects += [f"else/{os.path.splitext(f)[0]}" for f in os.listdir(cls.__ELSELIB_DIR) if f.endswith(".pd")]
+        pd_objects.extend(cls.__PD_CLASSES.keys())
         return pd_objects
 
     @classmethod
-    def __get_hv_args(clazz, pd_path):
+    def __get_hv_args(cls, pd_path: str) -> OrderedDict:
         """ Pre-parse the file for Heavy arguments, such that they are available
             as soon as a graph is created.
         """
         num_canvas = -1
         hv_arg_dict = OrderedDict()
-        hv_arg_list = None
+        hv_arg_list: Optional[List] = None
         with open(pd_path, "r") as f:
             for li in f:
                 if li.startswith("#N canvas"):
                     hv_arg_list = []
                     hv_arg_dict[li.rstrip(";\r\n")] = hv_arg_list
                     num_canvas += 1
-                elif "@hv_arg" in li:
+                elif "@hv_arg" in li and hv_arg_list is not None:
                     hv_arg_list.append(li.rstrip(";\r\n"))
                 elif li.startswith("#X restore"):
                     num_canvas -= 1
@@ -95,7 +101,7 @@ class PdParser:
         return hv_arg_dict
 
     @classmethod
-    def __get_pd_line(clazz, pd_path):
+    def __get_pd_line(cls, pd_path: str) -> Generator:
         concat = ""  # concatination state
         with open(pd_path, "r") as f:
             for li in f:
@@ -110,20 +116,20 @@ class PdParser:
                 else:
                     concat = (f'{concat} {li}') if len(concat) > 0 else f'{li}'
 
-    def add_absolute_search_directory(self, search_dir):
+    def add_absolute_search_directory(self, search_dir: str) -> bool:
         if os.path.isdir(search_dir):
             self.__search_paths.append(search_dir)
             return True
         else:
             return False
 
-    def add_relative_search_directory(self, search_dir):
+    def add_relative_search_directory(self, search_dir: str) -> bool:
         search_dir = os.path.abspath(os.path.join(
             self.__search_paths[0],
             search_dir))
         return self.add_absolute_search_directory(search_dir)
 
-    def find_abstraction_path(self, local_dir, abs_name):
+    def find_abstraction_path(self, local_dir: str, abs_name: str) -> Optional[str]:
         """ Finds the full path for an abstraction.
             Checks the local directory first, then all declared paths.
         """
@@ -143,7 +149,15 @@ class PdParser:
 
         return None
 
-    def graph_from_file(self, file_path, obj_args=None, pos_x=0, pos_y=0, is_root=True, pd_graph_class=PdGraph):
+    def graph_from_file(
+        self,
+        file_path: str,
+        obj_args: Optional[List] = None,
+        pos_x: int = 0,
+        pos_y: int = 0,
+        is_root: bool = True,
+        pd_graph_class: Type[PdGraph] = PdGraph
+    ) -> Any:
         """ Instantiate a PdGraph from a file.
             Note that obj_args does not include $0.
             @param pd_graph_class  The python class to handle specific graph types
@@ -154,8 +168,8 @@ class PdParser:
         if is_root:
             self.__search_paths.append(os.path.dirname(file_path))
 
-        file_hv_arg_dict = PdParser.__get_hv_args(file_path)
-        file_iterator = PdParser.__get_pd_line(file_path)
+        file_hv_arg_dict = self.__get_hv_args(file_path)
+        file_iterator = self.__get_pd_line(file_path)
         canvas_line = file_iterator.__next__()
 
         self.__DOLLAR_ZERO += 1  # increment $0
@@ -177,7 +191,7 @@ class PdParser:
             pd_graph_class)
 
         if is_root:
-            if g.get_notices()["errors"]:
+            if g.get_notices().has_error:
                 # return the graph early here as there are already errors and it is
                 # clearly invalid, avoids triggering unrelated errors in validation
                 return g
@@ -185,8 +199,18 @@ class PdParser:
 
         return g
 
-    def graph_from_canvas(self, file_iterator, file_hv_arg_dict, canvas_line, graph_args,
-                          pd_path, pos_x=0, pos_y=0, is_root=False, pd_graph_class=PdGraph):
+    def graph_from_canvas(
+        self,
+        file_iterator: Generator,
+        file_hv_arg_dict: Dict,
+        canvas_line: str,
+        graph_args: List,
+        pd_path: str,
+        pos_x: int = 0,
+        pos_y: int = 0,
+        is_root: bool = False,
+        pd_graph_class: Type[PdGraph] = PdGraph
+    ) -> Any:
         """ Instantiate a PdGraph from an existing canvas.
             Note that graph_args includes $0.
             @param file_hv_arg_dict  A dictionary containing all Heavy argument lines
@@ -194,17 +218,18 @@ class PdParser:
             @param canvas_line  The "#N canvas" which initiates this canvas.
             @param pd_graph_class  The python class to handle specific graph types
         """
-        obj_array = None  # an #A (table) object which is currently being parsed
+        obj_array: Optional[HeavyObject] = None  # an #A (table) object which is currently being parsed
 
         g = pd_graph_class(graph_args, pd_path, pos_x, pos_y)
 
+        remotes: Dict = {}
+
         # parse and add all Heavy arguments to the graph
         for li in file_hv_arg_dict[canvas_line]:
             line = li.split()
             assert line[4] == "@hv_arg"
             is_required = (line[9] == "true")
-            default_value = HeavyObject.force_arg_type(line[8], line[7]) \
-                if not is_required else None
+            default_value = HeavyObject.force_arg_type(line[8], line[7]) if not is_required else None
             g.add_hv_arg(
                 arg_index=int(line[5][2:]) - 1,  # strip off the leading "\$" and make index zero-based
                 name=line[6],
@@ -215,7 +240,7 @@ class PdParser:
         try:  # this try will capture any critical errors
             for li in file_iterator:
                 # remove width parameter
-                line = PdParser.__RE_WIDTH.sub("", li).split()
+                line = self.__RE_WIDTH.sub("", li).split()
 
                 if line[0] == "#N":
                     if line[1] == "canvas":
@@ -232,7 +257,7 @@ class PdParser:
                 elif line[0] == "#X":
                     if line[1] == "restore":
                         if len(line) > 5 and line[5] == "@hv_obj":
-                            obj_args = PdParser.__resolve_object_args(
+                            obj_args = self.__resolve_object_args(
                                 obj_type=line[6],
                                 obj_args=line[7:],
                                 graph=g,
@@ -255,23 +280,23 @@ class PdParser:
                             # are we restoring an array object?
                             # do some final sanity checks
                             if obj_array is not None:
-                                declared_size = obj_array.obj_args["size"]
-                                values_size = len(obj_array.obj_args["values"])
+                                declared_size = obj_array.obj_dict["size"]
+                                values_size = len(obj_array.obj_dict["values"])
                                 if declared_size != values_size:
                                     new_size = max(declared_size, values_size)
                                     obj_array.add_warning(
                                         "Table \"{0}\" was declared as having {1} values, "
                                         "but {2} were supplied. It will be resized to {3} "
                                         "values (any unsupplied values will be zeroed).".format(
-                                            obj_array.obj_args["name"],
+                                            obj_array.obj_dict["name"],
                                             declared_size,
                                             values_size,
                                             new_size))
-                                    obj_array.obj_args["size"] = new_size
+                                    obj_array.obj_dict["size"] = new_size
                                     if new_size < declared_size:
-                                        obj_array.obj_args["values"] = obj_args["values"][:new_size]
+                                        obj_array.obj_dict["values"] = obj_args["values"][:new_size]
                                     else:
-                                        obj_array.obj_args["values"].extend([0.0 for _ in
+                                        obj_array.obj_dict["values"].extend([0.0 for _ in
                                                                              range(new_size - declared_size)])
                                 obj_array = None  # done parsing the array
 
@@ -295,12 +320,12 @@ class PdParser:
                         if len(line) > 4:
                             obj_type = line[4]
                             # sometimes objects have $ arguments in them as well
-                            obj_type = PdParser.__resolve_object_args(
+                            obj_type = self.__resolve_object_args(
                                 obj_type=obj_type,
                                 obj_args=[obj_type],
                                 graph=g,
                                 is_root=is_root)[0]
-                            obj_args = PdParser.__resolve_object_args(
+                            obj_args = self.__resolve_object_args(
                                 obj_type=obj_type,
                                 obj_args=line[5:],
                                 graph=g,
@@ -328,31 +353,31 @@ class PdParser:
                                 is_root=False)
 
                         # is this object in lib/pd_converted?
-                        elif os.path.isfile(os.path.join(PdParser.__PDLIB_CONVERTED_DIR, f"{obj_type}.hv.json")):
+                        elif os.path.isfile(os.path.join(self.__PDLIB_CONVERTED_DIR, f"{obj_type}.hv.json")):
                             self.obj_counter[obj_type] += 1
-                            hv_path = os.path.join(PdParser.__PDLIB_CONVERTED_DIR, f"{obj_type}.hv.json")
+                            hv_path = os.path.join(self.__PDLIB_CONVERTED_DIR, f"{obj_type}.hv.json")
                             x = HeavyGraph(
                                 hv_path=hv_path,
                                 obj_args=obj_args,
                                 pos_x=int(line[2]), pos_y=int(line[3]))
 
                         # is this object in lib/heavy_converted?
-                        elif os.path.isfile(os.path.join(PdParser.__HVLIB_CONVERTED_DIR, f"{obj_type}.hv.json")):
+                        elif os.path.isfile(os.path.join(self.__HVLIB_CONVERTED_DIR, f"{obj_type}.hv.json")):
                             self.obj_counter[obj_type] += 1
-                            hv_path = os.path.join(PdParser.__HVLIB_CONVERTED_DIR, f"{obj_type}.hv.json")
+                            hv_path = os.path.join(self.__HVLIB_CONVERTED_DIR, f"{obj_type}.hv.json")
                             x = HeavyGraph(
                                 hv_path=hv_path,
                                 obj_args=obj_args,
                                 pos_x=int(line[2]), pos_y=int(line[3]))
 
                         # is this object in lib/pd?
-                        elif os.path.isfile(os.path.join(PdParser.__PDLIB_DIR, f"{obj_type}.pd")):
+                        elif os.path.isfile(os.path.join(self.__PDLIB_DIR, f"{obj_type}.pd")):
                             self.obj_counter[obj_type] += 1
-                            pdlib_path = os.path.join(PdParser.__PDLIB_DIR, f"{obj_type}.pd")
+                            pdlib_path = os.path.join(self.__PDLIB_DIR, f"{obj_type}.pd")
 
                             # mapping of pd/lib abstraction objects to classes
                             # for checking connection validity
-                            clazz = {
+                            pd_class = {
                                 "abs~": PdLibSignalGraph,
                                 "clip~": PdLibSignalGraph,
                                 "cos~": PdLibSignalGraph,
@@ -377,7 +402,7 @@ class PdParser:
                                 obj_args=obj_args,
                                 pos_x=int(line[2]), pos_y=int(line[3]),
                                 is_root=False,
-                                pd_graph_class=clazz)
+                                pd_graph_class=pd_class)
 
                             # register any object-specific warnings or errors
                             if obj_type in {"rzero~", "rzero_rev~", "czero~", "czero_rev~"}:
@@ -386,9 +411,29 @@ class PdParser:
                                     "Arguments and control connections are ignored.")
 
                         # is this object in lib/heavy?
-                        elif os.path.isfile(os.path.join(PdParser.__HVLIB_DIR, f"{obj_type}.pd")):
+                        elif os.path.isfile(os.path.join(self.__HVLIB_DIR, f"{obj_type}.pd")):
                             self.obj_counter[obj_type] += 1
-                            hvlib_path = os.path.join(PdParser.__HVLIB_DIR, f"{obj_type}.pd")
+                            hvlib_path = os.path.join(self.__HVLIB_DIR, f"{obj_type}.pd")
+                            x = self.graph_from_file(
+                                file_path=hvlib_path,
+                                obj_args=obj_args,
+                                pos_x=int(line[2]), pos_y=int(line[3]),
+                                is_root=False)
+
+                        # is this object in lib/else?
+                        elif os.path.isfile(os.path.join(self.__ELSELIB_DIR, f"{obj_type}.pd")):
+                            self.obj_counter[obj_type] += 1
+                            hvlib_path = os.path.join(self.__ELSELIB_DIR, f"{obj_type}.pd")
+                            x = self.graph_from_file(
+                                file_path=hvlib_path,
+                                obj_args=obj_args,
+                                pos_x=int(line[2]), pos_y=int(line[3]),
+                                is_root=False)
+
+                        # is this object in lib? (sub-directory)
+                        elif os.path.isfile(os.path.join(self.__LIB_DIR, f"{obj_type}.pd")):
+                            self.obj_counter[obj_type] += 1
+                            hvlib_path = os.path.join(self.__LIB_DIR, f"{obj_type}.pd")
                             x = self.graph_from_file(
                                 file_path=hvlib_path,
                                 obj_args=obj_args,
@@ -396,15 +441,15 @@ class PdParser:
                                 is_root=False)
 
                         # is this an object that must be programatically parsed?
-                        elif obj_type in PdParser.__PD_CLASSES:
+                        elif obj_type in self.__PD_CLASSES:
                             self.obj_counter[obj_type] += 1
-                            obj_class = PdParser.__PD_CLASSES[obj_type]
+                            obj_class = self.__PD_CLASSES[obj_type]
                             x = obj_class(
                                 obj_type,
                                 obj_args,
                                 pos_x=int(line[2]), pos_y=int(line[3]))
 
-                        elif PdParser.__is_float(obj_type):
+                        elif self.__is_float(obj_type):
                             # parse float literals
                             self.obj_counter["float"] += 1
                             x = HeavyObject(
@@ -427,7 +472,7 @@ class PdParser:
                     elif line[1] in {"floatatom", "symbolatom"}:
                         self.obj_counter[line[1]] += 1
                         x = self.graph_from_file(
-                            file_path=os.path.join(PdParser.__PDLIB_DIR, f"{line[1]}.pd"),
+                            file_path=os.path.join(self.__PDLIB_DIR, f"{line[1]}.pd"),
                             obj_args=[],
                             pos_x=int(line[2]), pos_y=int(line[3]),
                             is_root=False)
@@ -437,7 +482,7 @@ class PdParser:
                         assert obj_array is None, "#X array object is already being parsed."
                         # array names can have dollar arguments in them.
                         # ensure that they are resolved
-                        table_name = PdParser.__resolve_object_args(
+                        table_name = self.__resolve_object_args(
                             obj_type="array",
                             obj_args=[line[2]],
                             graph=g)[0]
@@ -453,11 +498,19 @@ class PdParser:
 
                     elif line[1] == "msg":
                         self.obj_counter["msg"] += 1
-                        g.add_object(PdMessageObject(
+                        msg = PdMessageObject(
                             obj_type="msg",
                             obj_args=[" ".join(line[4:])],
                             pos_x=int(line[2]),
-                            pos_y=int(line[3])))
+                            pos_y=int(line[3]))
+
+                        index = g.add_object(msg)
+
+                        if len(msg.obj_dict) > 0:
+                            remotes[index] = []
+
+                            for remote in msg.obj_dict['remote']:
+                                remotes[index].append(remote)
 
                     elif line[1] == "connect":
                         g.add_parsed_connection(
@@ -472,7 +525,9 @@ class PdParser:
                                 "[declare] objects are not supported in abstractions. "
                                 "They can only be in the root canvas.")
                         elif len(line) >= 4 and line[2] == "-path":
-                            did_add = self.add_relative_search_directory(line[3])
+                            pd_parent = Path(pd_path).parent
+                            pd_search = os.path.join(pd_parent, line[3])
+                            did_add = self.add_relative_search_directory(pd_search)
                             if not did_add:
                                 g.add_warning(
                                     f"\"{line[3]}\" is not a valid relative abstraction "
@@ -486,12 +541,17 @@ class PdParser:
 
                     elif line[1] == "coords":
                         pass  # don't do anything with this command
+                    elif line[1] == "f":
+                        # A line such as "#X f 123"
+                        # is found after a "#X restore" when the [pd afasdfasd]
+                        # object's box has been resized
+                        pass  # don't do anything with this command
 
                     else:
                         g.add_error(f"Don't know how to parse line: {' '.join(line)}")
 
-                elif line[0] == "#A":
-                    obj_array.obj_args["values"].extend([float(f) for f in line[2:]])
+                elif line[0] == "#A" and obj_array is not None:
+                    obj_array.obj_dict["values"].extend([float(f) for f in line[2:]])
 
                 else:
                     g.add_error(f"Don't know how to parse line: {' '.join(line)}")
@@ -506,10 +566,39 @@ class PdParser:
                 # Sometimes it's all that we have, so perhaps it's a good idea.
                 g.add_error(str(e), NotificationEnum.ERROR_EXCEPTION)
 
+        # parse remote messages
+        for index in remotes.keys():
+            first_msg = g.get_object(index)
+            conns = first_msg.get_inlet_connections()
+
+            for remote in remotes[index]:
+                self.obj_counter["msg"] += 1
+                msg = PdMessageObject('msg', [' '.join(msg for msg in remote['message'])])
+                msg_index = g.add_object(msg)
+
+                self.obj_counter["send"] += 1
+                send = PdSendObject('send', [remote['receiver']])
+                send_index = g.add_object(send)
+
+                # connect new message to upstream objects of first message
+                for conn in conns['0']:
+                    up_obj = conn.from_obj
+                    up_index = g.get_objects().index(up_obj)
+                    g.add_parsed_connection(up_index, 0, msg_index, 0)
+
+                g.add_parsed_connection(msg_index, 0, send_index, 0)
+
         return g
 
     @classmethod
-    def __resolve_object_args(clazz, obj_type, obj_args, graph, raise_on_failure=True, is_root=False):
+    def __resolve_object_args(
+        cls,
+        obj_type: str,
+        obj_args: List,
+        graph: PdGraph,
+        raise_on_failure: bool = True,
+        is_root: bool = False
+    ) -> Any:
         """ Resolve object arguments against the given graph arguments.
             By default this function raises an Exception if it cannot resolve an
             argument.
@@ -521,7 +610,7 @@ class PdParser:
         # TODO(mhroth): can this be done more elegantly?
         resolved_obj_args = list(obj_args)  # make a copy of the original obj_args
         for i, a in enumerate(obj_args):
-            for m in set(PdParser.__RE_DOLLAR.findall(a)):
+            for m in set(cls.__RE_DOLLAR.findall(a)):
                 x = int(m)  # the dollar index (i.e. $x)
                 if len(graph.obj_args) > x:
                     a = a.replace(fr"\${m}", str(graph.obj_args[x]))
@@ -592,7 +681,7 @@ class PdParser:
         __PD_CLASSES[o] = PdBinopObject
 
     @classmethod
-    def __is_float(clazz, x):
+    def __is_float(cls, x: int) -> bool:
         """ Returns True if the input can be converted to a float. False otherwise.
         """
         try:
